@@ -1,4 +1,3 @@
-#!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
 #
@@ -8,12 +7,11 @@
 # file LICENSE or http://www.opensource.org/licenses/mit-license.php.
 #
 
+import six
+
 from struct import pack, unpack
 
-try:
-    from cStringIO import StringIO
-except:
-    from StringIO import StringIO
+from .utils import StringIO
 
 from python_patterns.utils.decorators import Property
 
@@ -424,8 +422,30 @@ OPCODE_NAMES = {
 
 # ===----------------------------------------------------------------------===
 
-class ScriptOp(SerializableMixin):
-    def __init__(self, opcode=None, data=None, *args, **kwargs):
+class ScriptOp(SerializableMixin, six.binary_type):
+    def __new__(cls, opcode=None, data=None, *args, **kwargs):
+        # If just data is specified, then we automatically fill-in the opcode
+        # as required to push the data on the stack.
+        if opcode is None and data:
+            len_ = len(data)
+            if 0 < len_ < OP_PUSHDATA1:
+                opcode = len_
+            elif OP_PUSHDATA1 <= len_ <= 0xff:
+                opcode = OP_PUSHDATA1
+            elif 0xff < len_ < 0xffff:
+                opcode = OP_PUSHDATA2
+            elif 0xffff < len_ < 0xffffffff:
+                opcode = OP_PUSHDATA4
+            else:
+                if not len_:
+                    raise TypeError(u"cannot push empty string")
+                else:
+                    raise TypeError(u"string exceeds maximum length")
+
+        # The push opcode encodes the length of the data to be pushed, or
+        # the size of the length which directly follows. In either case we
+        # need to make sure the actual data length meets or does not exceed
+        # this limit.
         if opcode==OP_INVALIDOPCODE:
             raise ValueError(u"invalid opcode")
         elif opcode>0 and opcode<OP_PUSHDATA1 and opcode!=len(data):
@@ -434,21 +454,36 @@ class ScriptOp(SerializableMixin):
               opcode==OP_PUSHDATA2 and len(data)>0xffff or
               opcode==OP_PUSHDATA4 and len(data)>0xffffffff):
             raise ValueError(u"data length exceeds serialization limit")
-        super(ScriptOp, self).__init__(*args, **kwargs)
-        self.opcode = opcode
-        self.data = data
+
+        result = pack('<B', opcode)
+        if opcode==OP_PUSHDATA1:
+            result += pack('<B', len(data))
+        elif opcode==OP_PUSHDATA2:
+            result += pack('<H', len(data))
+        elif opcode==OP_PUSHDATA4:
+            result += pack('<I', len(data))
+        if opcode>0 and opcode<=OP_PUSHDATA4:
+            result += data
+
+        return super(ScriptOp, cls).__new__(cls, result, *args, **kwargs)
+
+    @Property
+    def opcode():
+        def fget(self):
+            return unpack('>B', self[:1])[0]
+        return locals()
+
+    @Property
+    def data():
+        def fget(self):
+            if self.opcode <= OP_PUSHDATA4:
+                return self[1:]
+            else:
+                return None
+        return locals()
 
     def serialize(self):
-        result  = chr(self.opcode)
-        if self.opcode==OP_PUSHDATA1:
-            result += pack('<B', len(self.data))
-        elif self.opcode==OP_PUSHDATA2:
-            result += pack('<H', len(self.data))
-        elif self.opcode==OP_PUSHDATA4:
-            result += pack('<I', len(self.data))
-        if self.opcode>0 and self.opcode<=OP_PUSHDATA4:
-            result += self.data
-        return result
+        return self
     @classmethod
     def deserialize(cls, file_):
         opcode = file_.read(1)
@@ -469,7 +504,7 @@ class ScriptOp(SerializableMixin):
         else:
             data = ''
         if len(data) != datalen:
-            raise ValueError(u"unespected end-of-file in data string")
+            raise ValueError(u"unexpected end-of-file in data string")
 
         if opcode<=OP_PUSHDATA4:
             return cls(opcode, data)
@@ -541,14 +576,6 @@ class ScriptOp(SerializableMixin):
                 self.data = data
         return locals()
 
-    def __eq__(self, other):
-        if isinstance(other, str):
-            return self.serialize() == other
-        if self.opcode != other.opcode:
-            return False
-        if self.opcode<=OP_PUSHDATA4 and self.data != other.data:
-            return False
-        return True
     def __repr__(self):
         if self.opcode>0 and self.opcode<=OP_PUSHDATA4:
             return ''.join(['0x', self.data.encode('hex')])
@@ -558,6 +585,8 @@ class ScriptOp(SerializableMixin):
             else:
                 return "OP_UNKNOWN"
 
+# ===----------------------------------------------------------------------===
+
 class Script(SerializableMixin, tuple):
     def __new__(cls, *args, **kwargs):
         if len(args)==1 and isinstance(args[0], str) and not kwargs:
@@ -585,12 +614,156 @@ class Script(SerializableMixin, tuple):
 
 # ===----------------------------------------------------------------------===
 
+from .crypto import VerifyingKey
+from .defaults import CLIENT_VERSION
+from .serialize import SER_DISK, serialize_varint, deserialize_varchar
+from .utils import StringIO
+
+class ScriptPickler(object):
+    """Compact serializer for scripts.
+
+    It detects common cases and encodes them much more efficiently.
+    3 special cases are defined:
+      * Pay to pubkey hash (encoded as 21 bytes)
+      * Pay to script hash (encoded as 21 bytes)
+      * Pay to pubkey starting with 0x02, 0x03 or 0x04 (encoded as 33 bytes)
+
+    Other scripts up to 121 bytes require 1 byte + script length. Above
+    that, scripts up to 16505 bytes require 2 bytes + script length."""
+    def __init__(self, file=None, protocol=SER_DISK, version=CLIENT_VERSION,
+                 *args, **kwargs):
+        if file is None: file = StringIO()
+        super(ScriptPickler, self).__init__(*args, **kwargs)
+        self._file = file
+        self._protocol = protocol
+        self._version = version
+
+    @staticmethod
+    def _dump(script, file_, protocol, version):
+        if hasattr(script, 'serialize'):
+            script = script.serialize()
+            script = deserialize_varchar(StringIO(script))
+        script_len = len(script)
+        if 23 == script_len and all(
+            script[k:k+1] == six.int2byte(v)
+                    for k,v in six.iteritems({
+                        0:  OP_HASH160,
+                        1:  20,
+                        22: OP_EQUAL})):
+            str_ = b''.join([b'\x01', script[2:22]])
+        elif 25 == script_len and all(
+                script[k:k+1] == six.int2byte(v)
+                    for k,v in six.iteritems({
+                        0:  OP_DUP,
+                        1:  OP_HASH160,
+                        2:  20,
+                        23: OP_EQUALVERIFY,
+                        24: OP_CHECKSIG})):
+            str_ = b''.join([b'\x00', script[3:23]])
+        elif 35 == script_len and script[1:2] in map(six.int2byte, [2, 3]) and all(
+                script[k:k+1] == six.int2byte(v)
+                    for k,v in six.iteritems({
+                        0:  33,
+                        34: OP_CHECKSIG})):
+            str_ = script[1:34]
+        elif 67 == script_len and all(
+                script[k:k+1] == six.int2byte(v)
+                    for k,v in six.iteritems({
+                        0:  65,
+                        1:  4,
+                        66: OP_CHECKSIG})):
+            str_ = b''.join([six.int2byte(0x04 | ord(script[65:66])&0x01), script[2:34]])
+        else:
+            str_ = b''.join([serialize_varint(script_len + 0x06), script])
+        file_.write(str_)
+
+    @staticmethod
+    def _load(file_, protocol, version):
+        size = unpack("<B", file_.read(1))[0]
+        if size == 0:
+            hash_ = file_.read(20)
+            return b''.join([
+                six.int2byte(25),
+                six.int2byte(OP_DUP),
+                six.int2byte(OP_HASH160),
+                six.int2byte(20), hash_,
+                six.int2byte(OP_EQUALVERIFY),
+                six.int2byte(OP_CHECKSIG),
+            ])
+        elif size == 1:
+            hash_ = file_.read(20)
+            return b''.join([
+                six.int2byte(23),
+                six.int2byte(OP_HASH160),
+                six.int2byte(20), hash_,
+                six.int2byte(OP_EQUAL),
+            ])
+        elif size in (2, 3):
+            compressed_key = file_.read(32)
+            return b''.join([
+                six.int2byte(35),
+                six.int2byte(33),
+                six.int2byte(size),
+                compressed_key,
+                six.int2byte(OP_CHECKSIG),
+            ])
+        elif size in (4, 5):
+            verifying_key = VerifyingKey.deserialize(StringIO(
+                b''.join([six.int2byte(size-2), file_.read(32)])))
+            verifying_key.compressed = False
+            return b''.join([
+                six.int2byte(67),
+                six.int2byte(65),
+                verifying_key.serialize(),
+                six.int2byte(OP_CHECKSIG),
+            ])
+        elif size == 0xfd:
+            size = unpack("<H", file_.read(2))[0]
+        elif size == 0xfe:
+            size = unpack("<I", file_.read(4))[0]
+        elif size == 0xff:
+            size = unpack("<Q", file_.read(8))[0]
+        size = size - 6
+        script = file_.read(size)
+        return serialize_varchar(script)
+
+    def get_script_class(self):
+        return getattr(self, 'script_class', Script)
+
+    def dump(self, script):
+        "Write a compressed representation of script to the Pickler's file object."
+        self._dump(script, self._file, self._protocol, self._version)
+
+    def dumps(self, script):
+        "Return a compressed representation of script as a binary string."
+        string = StringIO()
+        self._dump(script, string, self._protocol, self._version)
+        return string.getvalue()
+
+    def load(self):
+        "Read and decompress a compact script from the Pickler's file object."
+        script_class = self.get_script_class()
+        script = self._load(self._file, self._protocol, self._version)
+        return script_class.deserialize(StringIO(script))
+
+    def loads(self, string):
+        "Decompress the passed-in compact script and return the result."
+        script_class = self.get_script_class()
+        script = self._load(StringIO(string), self._protocol, self._version)
+        return script_class.deserialize(StringIO(script))
+
+ScriptUnpickler = ScriptPickler
+
+# ===----------------------------------------------------------------------===
+
 __all__ = OPCODE_NAMES.values() + [
     'OP_TRUE',
     'OP_FALSE',
     'OP_INVALIDOPCODE',
     'Script',
     'ScriptOp',
+    'ScriptPickler',
+    'ScriptUnpickler',
 ]
 
 #
